################
# Dependencies #
################
from flask import request
from flask.views import View


####################
# Controller Class #
####################
##
# @desc Controller class to control views based on the requested method 
##
class Controller(View):

    ##
    # @desc Flask dispatch_request method -- Generates Pluggable Views
    ##
    def dispatch_request(self, *class_args, **class_kwargs):
        # Check the requested methods then return the related view function
        # The 'POST' request
        if request.method == 'POST':
            return self.post(*class_args, **class_kwargs)

        # The 'GET' request
        elif request.method == 'GET':
            return self.get(*class_args, **class_kwargs)

        # The 'PUT' request
        elif request.method == 'PUT':
            return self.put(*class_args, **class_kwargs)

        # The 'DELETE' request
        elif request.method == 'DELETE':
            return self.delete(*class_args, **class_kwargs)


    ##
    # @desc get method placeholder -- To handle the 'GET' requests
    # 
    # @return any
    ##
    def post(self):
        return 'POST Method'


    ##
    # @desc get method placeholder -- To handle the 'GET' requests
    # 
    # @return any
    ##
    def get(self):
        return 'GET Method'
     

    ##
    # @desc get method placeholder -- To handle the 'GET' requests
    # 
    # @return any
    ##
    def put(self):
        return 'PUT Method'
     

    ##
    # @desc get method placeholder -- To handle the 'DELETE' requests
    # 
    # @return any
    ##
    def delete(self):
        return 'DELETE Method'




################
# Dependencies #
################
import sys
import pathlib
from flask import render_template


################
# Aurora Class #
################
##
# @desc Aurora View to render the specified app view 
##
class View():
    
    ##
    # @desc Constructor method
    ##
    def __init__(self):
        pass


    ##
    # @desc render the view template
    # 
    # @param view: str -- The view template (ex. 'index.html')
    # 
    # @var caller: str -- The caller file path
    # @var app: str -- The caller app name
    # @var view_path: str -- The view template path
    # 
    # @return text/html: str -- The rendered template
    ##
    def render(self, view:str, *class_args, **class_kwargs):
        # Check the view
        # User inserted an absolute view template
        if '/' in view:
            view_path = view

        # User inserted a relative view template
        else:
            # Inspect the app name from caller file
            caller = sys._getframe().f_back.f_code.co_filename
            app = pathlib.PurePath(caller).parent.name
            view_path = f'{app}/{view}'

        # Render the view template      
        return render_template(view_path, *class_args, **class_kwargs)


    ##
    # @desc render the error view template
    # 
    # @param code: int -- The error http code (ex. 400)
    # @param view: str -- The view template (ex. 'index.html')
    # 
    # @var caller: str -- The caller file path
    # @var app: str -- The caller app name
    # @var view_path: str -- The view template path
    # 
    # @return tuple[str, int] -- str: The rendered template -- int: HTTP error code
    ##
    def render_error(self, code:int, view:str, *class_args, **class_kwargs):
        # Check the view
        # User inserted an absolute view template
        if '/' in view:
            view_path = view

        # User inserted a relative view template
        else:
            # Inspect the app name from caller file
            caller = sys._getframe().f_back.f_code.co_filename
            app = pathlib.PurePath(caller).parent.name
            view_path = f'{app}/{view}'

        # Render the view template      
        return render_template(view_path, *class_args, **class_kwargs), int(code)




# Dependencies
from .Aurora import Aurora
from .Controller import Controller
from .View import View as Template

try:
    from .Forms import Forms
    from .Model import Model
except:
    pass

View = Template().render
Error = Template().render_error


# Descriptions
__version__ = '0.5.3'
__author__ = 'Hemin Satya <https://github.com/heminsatya>'



################
# Dependencies #
################
import os
import platform
import importlib
import click
from datetime import datetime
from .helpers import *
from .SQL import Database

# Instantiate the Database class
db = Database()

# Aurora path
aurora_path = os.path.dirname(__file__)

# App path (the caller)
app_path = os.getcwd()

# Check platform system
# Windows
if platform.system() == 'Windows':
    url_div = '\\'

# Linux, Mac
else:
    url_div = '/'

# Fetch statics
config = importlib.import_module('config')
statics = getattr(config, "STATICS")

# Fetch registered apps
apps_module = importlib.import_module('_apps')
apps = getattr(apps_module, "apps")

# Fetch registered models
models_module = importlib.import_module('models._models')
models = getattr(models_module, "models")

# The database (file)
database = getattr(config, "DB_CONFIG")['database']


#############
# CLI Class #
#############
##
# @desc CLI class for manage the root app and its child apps
##
class CLI:
    ##
    # @desc Constructor method
    ##
    def __init__(self):

        # Try to run the CLI application
        try:
            # Call the group method
            self.group()

            # Call the cli method
            self.cli()

        # Handle errorr
        except NameError as err:
            raise Exception(err)


    ##
    # @desc group method for grouping commands together
    ##
    def group(self):
        self.cli.add_command(self.create_app)
        self.cli.add_command(self.delete_app)
        self.cli.add_command(self.create_controller)
        self.cli.add_command(self.delete_controller)
        self.cli.add_command(self.create_view)
        self.cli.add_command(self.delete_view)
        self.cli.add_command(self.create_model)
        self.cli.add_command(self.delete_model)
        self.cli.add_command(self.create_form)
        self.cli.add_command(self.delete_form)
        self.cli.add_command(self.check_db)
        self.cli.add_command(self.init_db)
        self.cli.add_command(self.migrate_db)


    ##
    # @desc cli method for grouping other methods
    ##
    @click.group()
    def cli():
        pass


    ##
    # @desc create_app method for creating new child apps
    # 
    # @var name: str -- The app name
    # @var url: str -- The app base URL
    ##
    @click.command()
    def create_app():
        # Prompt for app name
        while True:
            name = input("App Name: ")

            # Check app name
            if app_name(name)['result']:
                # App name is taken
                if name in apps:
                    print(f'The "{name}" is already registered. Try another name.')

                # App name is OK
                else:
                    break

            # Print the error
            else:
                print(app_name(name)['message'])

        # Prompt for base URL
        while True:
            url = input("Base URL: ")

            # Check base URL
            if base_url(url)['result']:
                # Base URL is taken
                if url in apps.values():
                    print(f'The "{url}" is already registered. Try another URL.')

                # Base URL is OK
                else:
                    break

            # Print the error
            else:
                print(base_url(url)['message'])
        
        # Begin the process
        try:
            print('Creating the new app...')

            # Create folders: (controllers, forms, statics, views)
            make_dir(f'{app_path + url_div}controllers{url_div + name}')
            make_dir(f'{app_path + url_div}forms{url_div + name}')
            make_dir(f'{app_path + url_div + statics + url_div + name}')
            make_dir(f'{app_path + url_div}views{url_div + name}')

            # Handle controllers blueprint (copy, unzip, delete)
            controllers_blueprint= f'{aurora_path + url_div}blueprints{url_div}controllers.zip'
            controllers_file = f'{app_path + url_div}controllers{url_div + name + url_div}controllers.zip'
            controllers_dir = f'{app_path + url_div}controllers{url_div + name + url_div}'
            copy_file(controllers_blueprint, controllers_file)
            unzip_file(controllers_file, controllers_dir)
            delete_file(controllers_file)

            # Handle forms blueprint (copy, unzip, delete)
            forms_blueprint= f'{aurora_path + url_div}blueprints{url_div}forms.zip'
            forms_file = f'{app_path + url_div}forms{url_div + name + url_div}forms.zip'
            forms_dir = f'{app_path + url_div}forms{url_div + name + url_div}'
            copy_file(forms_blueprint, forms_file)
            unzip_file(forms_file, forms_dir)
            delete_file(forms_file)

            # Handle statics blueprint (copy, unzip, delete)
            statics_blueprint= f'{aurora_path + url_div}blueprints{url_div}statics.zip'
            statics_file = f'{app_path + url_div + statics + url_div + name + url_div}statics.zip'
            statics_dir = f'{app_path + url_div + statics + url_div + name + url_div}'
            copy_file(statics_blueprint, statics_file)
            unzip_file(statics_file, statics_dir)
            delete_file(statics_file)

            # Handle views blueprint (copy, unzip, delete)
            views_blueprint= f'{aurora_path + url_div}blueprints{url_div}views.zip'
            views_file = f'{app_path + url_div}views{url_div + name + url_div}views.zip'
            views_dir = f'{app_path + url_div}views{url_div + name + url_div}'
            copy_file(views_blueprint, views_file)
            unzip_file(views_file, views_dir)
            delete_file(views_file)

            # Update layout.html
            replace_file_string(f'{app_path + url_div}views{url_div + name + url_div}layout.html', 'app_name', name)

            # Update _apps.py
            new_line = f'''    '{name}': '{url}',\n'''
            new_line += '''}#do-not-change-me'''
            replace_file_line(file_path=f'{app_path + url_div}_apps.py', old_line='}#do-not-change-me', new_line=new_line)

            # print the message
            print('The new app created successfully!')

        # Handle errors
        except NameError as err:
            print(err)


    ##
    # @desc delete_app method for removing child apps
    ##
    @click.command()
    def delete_app():
        # Prompt for app name
        while True:
            name = input("App Name: ")

            # Check app name
            if app_name(name)['result']:
                # App not exists
                if not name.lower() in apps:
                    print(f'The "{name}" doesn\'t exist!')

                # App exists
                else:
                    break

            # Print the error
            else:
                print(app_name(name)['message'])
        
        # Alert the user for data loss
        alert = '''You will loose the following data perminantly:\n'''
        alert += f'''  {app_path + url_div}controllers{url_div + name + url_div}*\n'''
        alert += f'''  {app_path + url_div}forms{url_div + name + url_div}*\n'''
        alert += f'''  {app_path + url_div + statics + url_div + name + url_div}*\n'''
        alert += f'''  {app_path + url_div}views{url_div + name + url_div}*\n'''
        
        # Print the alert
        print(alert)

        # Prompt the user for confirmation
        confirm = input("Do you want to continue? (yes/no) ")
            
        # Confirmed
        if confirm.lower() == 'yes':
            print('Removing the app...')

            # Begin the process
            try:
                # Delete the app folders
                delete_dir(f'{app_path + url_div}controllers{url_div + name + url_div}')
                delete_dir(f'{app_path + url_div}forms{url_div + name + url_div}')
                delete_dir(f'{app_path + url_div + statics + url_div + name + url_div}')
                delete_dir(f'{app_path + url_div}views{url_div + name + url_div}')

                # Update the _apps.py
                url = apps[name]
                old_line = f"""'{name}': '{url}',"""
                replace_file_line(file_path=f'{app_path + url_div}_apps.py', old_line=old_line, new_line='')
                
                print('App deleted successfully')

            # Handle errors
            except NameError as err:
                print(err)

        # Rejected
        else:
            print('The operation canceled!')
            exit()


    ##
    # @desc create_controller method for creating new controller
    ##
    @click.command()
    def create_controller():
        # Prompt for app name
        while True:
            name = input("App Name: ")

            # Check app name
            if app_name(name)['result']:
                # App not exists
                if not name.lower() in apps:
                    print(f'The "{name}" doesn\'t exist!')

                # App exists
                else:
                    break

            # Print the error
            else:
                print(app_name(name)['message'])

        # Controllers info
        module = importlib.import_module(f'controllers.{name}._controllers')
        controllers = getattr(module, 'controllers')

        # Controller name
        break_loop = False
        while True:
            # Check break loop from for loop
            if break_loop:
                break
            
            # Prompt for controller name
            ctrl = input("Controller Name: ")

            # Check controller name
            if controller_name(ctrl)['result']:
                if len(controllers) == 0:
                    break

                # Controller already exists
                for controller in controllers:
                    if ctrl in controller:
                        print(f'The "{ctrl}" already exists!')
                        break

                    # Controller not exists
                    else:
                        break_loop = True
                        break

            # Print the error
            else:
                print(controller_name(ctrl)['message'])

        # Prompt for controller url
        break_loop = False
        while True:
            # Check break loop from for loop
            if break_loop:
                break
            
            # Prompt for controller url
            if len(controllers) == 0:
                url = input("Controller URL (optional): ")
            else:
                url = input("Controller URL: ")

            # Check controller url
            if controller_url(url)['result']:
                if len(controllers) == 0:
                    break

                # Controller url already exists
                for controller in controllers:
                    if url in controller:
                        print(f'The "{url}" already exists!')
                        break

                    # Controller url not exists
                    else:
                        break_loop = True
                        break

            # Print the error
            else:
                print(controller_url(url)['message'])

        # Prompt for optional methods
        while True:
            methods = input("Methods (optional): ").upper()

            # Remove spaces from methods
            methods = delete_chars(methods, ' ')
            
            # Convert methods to list 
            if methods:
                methods = methods.split(',')

            # Sort methods
            methods = sorted(methods)

            # Check methods
            if not controller_methods(methods)['result']:
                print(controller_methods(methods)['message'])

            # Break the loop
            else:
                break
        
        # Controller methods
        ctrl_methods = ''

        # POST method
        if 'POST' in methods:
            ctrl_methods += """    # POST Method\n"""
            ctrl_methods += """    def post(self):\n"""
            ctrl_methods += """        pass\n\n"""

        # GET method
        if 'GET' in methods:
            ctrl_methods += """    # GET Method\n"""
            ctrl_methods += """    def get(self):\n"""
            ctrl_methods += """        return 'Page content...'\n\n"""

        # PUT method
        if 'PUT' in methods:
            ctrl_methods += """    # PUT Method\n"""
            ctrl_methods += """    def put(self):\n"""
            ctrl_methods += """        pass\n\n"""

        # DELETE method
        if 'DELETE' in methods:
            ctrl_methods += """    # DELETE Method\n"""
            ctrl_methods += """    def delete(self):\n"""
            ctrl_methods += """        pass\n\n"""

        # Default method (GET)
        if not methods:
            ctrl_methods += """    # GET Method\n"""
            ctrl_methods += """    def get(self):\n"""
            ctrl_methods += """        return 'Page content...'\n\n"""
        
        # try the process
        try:
            # Create controller
            print('Creating the controller...')
            
            # Controller blueprint
            controller_blueprint = f'{aurora_path + url_div}blueprints{url_div}controller.zip'
            controller_file = f'{app_path + url_div}controllers{url_div + name + url_div}controller.zip'
            controller_dir = f'{app_path + url_div}controllers{url_div + name + url_div}'

            # Copy, unzip, delete blueprint
            copy_file(controller_blueprint, controller_file)
            unzip_file(controller_file, controller_dir)
            delete_file(controller_file)

            # Rename _controller.py
            controller_old = f'{app_path + url_div}controllers{url_div + name + url_div}_controller.py'
            controller_new = f'{app_path + url_div}controllers{url_div + name + url_div + ctrl}.py'
            rename_file(controller_old, controller_new)

            # Update new controller
            replace_file_string(controller_new, 'ControllerName', ctrl)
            replace_file_line(controller_new, '...', ctrl_methods)

            # Update _controllers.py
            controllers_file = f'{app_path + url_div}controllers{url_div + name + url_div}_controllers.py'

            if methods:
                controller_data = f"""   ('{ctrl}', '{url}', {methods}),\n"""
            else:
                controller_data = f"""   ('{ctrl}', '{url}'),\n"""

            controller_data += """]#do-not-change-me"""

            replace_file_line(controllers_file, ']#do-not-change-me', controller_data)

            # Print result
            print('The new controller created successfuly!')

        # Handle errors
        except NameError as err:
            print(err)


    ##
    # @desc delete_controller method for removing an existing controller
    ##
    @click.command()
    def delete_controller():
        # Prompt for app name
        while True:
            name = input("App Name: ")

            # Check app name
            if app_name(name)['result']:
                # App not exists
                if not name.lower() in apps:
                    print(f'The "{name}" doesn\'t exist!')

                # App exists
                else:
                    break

            # Print the error
            else:
                print(app_name(name)['message'])

        # Controllers info
        module = importlib.import_module(f'controllers.{name}._controllers')
        controllers = getattr(module, 'controllers')

        # App controllers
        if len(controllers) == 0:
            print(f'No controllers found for the {name}!')
            exit()

        # Controller name
        break_loop = False
        while True:
            # Check break loop from for loop
            if break_loop:
                break
            
            # Prompt for controller name
            ctrl = input("Controller Name: ")

            # Check controller name
            if controller_name(ctrl)['result']:
                i = 1
                for controller in controllers:
                    # Controller exists
                    if ctrl in controller:
                        break_loop = True
                        break

                    # Controller not exists
                    else:
                        if len(controllers) == i:
                            print(f'The "{ctrl}" does\'nt exist!')

                    i += 1

            # Print the error
            else:
                print(controller_name(ctrl)['message'])
        
        # Alert the user for data loss
        alert = '''You will loose the following data perminantly:\n'''
        alert += f'''  {app_path + url_div}controllers{url_div + name + url_div + ctrl}.py\n'''
        
        # Print the alert
        print(alert)

        # Prompt the user for confirmation
        confirm = input("Do you want to continue? (yes/no) ")
            
        # Confirmed
        if confirm.lower() == 'yes':
            print('Removing the controller...')

            # Begin the process
            try:
                # Delete the controller module
                delete_file(f'{app_path + url_div}controllers{url_div + name + url_div + ctrl}.py')

                # Update _controllers.py
                controllers_file = f'{app_path + url_div}controllers{url_div + name + url_div}_controllers.py'
                replace_file_line(controllers_file, f"('{ctrl}',", '')
                
                print('Controller deleted successfully')

            # Handle errors
            except NameError as err:
                print(err)

        # Rejected
        else:
            print('The operation canceled!')
            exit()


    ##
    # @desc create_view method for creating new view
    ##
    @click.command()
    def create_view():
        # Prompt for app name
        while True:
            name = input("App Name: ")

            # Check app name
            if app_name(name)['result']:
                # App not exists
                if not name.lower() in apps:
                    print(f'The "{name}" doesn\'t exist!')

                # App exists
                else:
                    break

            # Print the error
            else:
                print(app_name(name)['message'])

        # Prompt for view name
        while True:
            view = input("View Name: ")

            # Check the view name
            if view_name(view)['result']:
                view_file = f'{app_path + url_div}views{url_div + name + url_div + view}.html'

                # View exists
                if os.path.exists(view_file):
                    print(f'The "{view}" is already exists!')

                # View not exists
                else:
                    break

            # Print the error
            else:
                print(view_name(view)['message'])

        # Begin the process
        try:
            print('Creating the new view...')
            
            # View blueprint
            view_blueprint = f'{aurora_path + url_div}blueprints{url_div}view.zip'
            view_file = f'{app_path + url_div}views{url_div + name + url_div}view.zip'
            view_dir = f'{app_path + url_div}views{url_div + name + url_div}'
            
            # Copy the view blueprint
            copy_file(view_blueprint, view_file)

            # Unzip the view blueprint
            unzip_file(view_file, view_dir)

            # Delete the view zip file
            delete_file(view_file)

            # Rename the view blueprint
            old_view = f'{app_path + url_div}views{url_div + name + url_div}_view.html'
            new_view = f'{app_path + url_div}views{url_div + name + url_div + view}.html'
            rename_file(old_view, new_view)

            # update the view blue print
            replace_file_string(new_view, 'app_name', name)

            # Print the result
            print('The new view created successfully!')
        
        # Handle errors
        except NameError as err:
            print(err)


    ##
    # @desc delete_view method for removing an existing view
    ##
    @click.command()
    def delete_view():
        # Prompt for app name
        while True:
            name = input("App Name: ")

            # Check app name
            if app_name(name)['result']:
                # App not exists
                if not name.lower() in apps:
                    print(f'The "{name}" doesn\'t exist!')

                # App exists
                else:
                    break

            # Print the error
            else:
                print(app_name(name)['message'])

        # Prompt for view name
        while True:
            view = input("View Name: ")

            # Check the view name
            if view_name(view)['result']:
                view_file = f'{app_path + url_div}views{url_div + name + url_div + view}.html'

                # View not exists
                if not os.path.exists(view_file):
                    print(f'The "{view}" deos\'nt exist!')

                # View exists
                else:
                    break

            # Print the error
            else:
                print(view_name(view)['message'])
        
        # Alert the user for data loss
        alert = '''You will loose the following data perminantly:\n'''
        alert += f'''  {app_path + url_div}views{url_div + name + url_div + view}.html\n'''
        
        # Print the alert
        print(alert)

        # Prompt the user for confirmation
        confirm = input("Do you want to continue? (yes/no) ")
            
        # Confirmed
        if confirm.lower() == 'yes':
            # Begin the process
            try:
                print('Removing the view...')

                # Delete the view file
                view_file = f'{app_path + url_div}views{url_div + name + url_div + view}.html'
                delete_file(view_file)

                # Print the result
                print('The view deleted successfully!')

            # Handle errors
            except NameError as err:
                print(err)

        # Rejected
        else:
            print('The operation canceled!')
            exit()


    ##
    # @desc create_model method for creating new model
    ##
    @click.command()
    def create_model():
        # Prompt for model name
        while True:
            model = input("Model Name: ")

            # Check model name
            if model_name(model)['result']:
                # Model exists
                if model in models:
                    print(f'The "{model}" already exist!')

                # Model not exists
                else:
                    break

            # Print the error
            else:
                print(model_name(model)['message'])

        # Begin the process
        try:
            print('Creating the new model...')

            # Model blueprint
            model_blueprint = f'{aurora_path + url_div}blueprints{url_div}model.zip'
            model_file = f'{app_path + url_div}models{url_div}model.zip'
            model_dir = f'{app_path + url_div}models{url_div}'
            
            # Copy the model blueprint
            copy_file(model_blueprint, model_file)

            # Unzip the model blueprint
            unzip_file(model_file, model_dir)

            # Delete the model zip file
            delete_file(model_file)

            # Rename the _model.py
            old_model = f'{app_path + url_div}models{url_div}_model.py'
            new_model = f'{app_path + url_div}models{url_div + model}.py'
            rename_file(old_model, new_model)

            # Update the model blue print
            replace_file_string(new_model, 'ModelName', model)
            replace_file_string(new_model, '_table_name', snake_case(model))

            # Update the _models.py
            models_file = f'{app_path + url_div}models{url_div}_models.py'

            models_data = f"""    '{model}',\n"""
            models_data += """)#do-not-change-me"""

            replace_file_line(models_file, ')#do-not-change-me', models_data)

            # Update the __init__.py
            init_file = f'{app_path + url_div}models{url_div}__init__.py'

            if len(models) == 0:
                replace_file_line(init_file, '...', '')

            init_data = f"""    from .{model} import {model}\n"""
            init_data += """    #do-not-change-me\n"""

            replace_file_line(init_file, '#do-not-change-me', init_data)

            # Print the result
            print('The new model created successfully!')
        
        # Handle errors
        except NameError as err:
            print(err)


    ##
    # @desc delete_model method for removing an existing model
    ##
    @click.command()
    def delete_model():
        # Prompt for model name
        while True:
            model = input("Model Name: ")

            # Check model name
            if model_name(model)['result']:
                # Model not exists
                if not model in models:
                    print(f'The "{model}" doesn\'t exist!')

                # Model exists
                else:
                    break

            # Print the error
            else:
                print(model_name(model)['message'])
        
        # Alert the user for data loss
        alert = '''You will loose the following data perminantly:\n'''
        alert += f'''  {app_path + url_div}models{url_div + model}.py\n'''
        
        # Print the alert
        print(alert)

        # Prompt the user for confirmation
        confirm = input("Do you want to continue? (yes/no) ")
            
        # Confirmed
        if confirm.lower() == 'yes':
            # Begin the process
            try:
                print('Removing the model...')

                # Delete the model module
                model_file = f'{app_path + url_div}models{url_div + model}.py'
                delete_file(model_file)

                # Update the _models.py
                models_file = f'{app_path + url_div}models{url_div}_models.py'
                replace_file_line(models_file, f"'{model}',", '')

                # Update the __init__.py
                init_file = f'{app_path + url_div}models{url_div}__init__.py'

                if len(models) == 1:
                    replace_file_line(init_file, f'from .{model} import {model}', '    ...\n')
                else:
                    replace_file_line(init_file, f'from .{model} import {model}', '')

                # Print the result
                print('The model deleted successfully!')

            # Handle errors
            except NameError as err:
                print(err)

        # Rejected
        else:
            print('The operation canceled!')
            exit()


    ##
    # @desc create_form method for creating new form
    ##
    @click.command()
    def create_form():
        # Prompt for app name
        while True:
            name = input("App Name: ")

            # Check app name
            if app_name(name)['result']:
                # App not exists
                if not name.lower() in apps:
                    print(f'The "{name}" doesn\'t exist!')

                # App exists
                else:
                    break

            # Print the error
            else:
                print(app_name(name)['message'])

        # Fetch registered forms for the app
        forms_module = importlib.import_module(f'forms.{name}._forms')
        forms = getattr(forms_module, "forms")
        
        # Prompt for form name
        while True:
            form = input("Form Name: ")

            # Check form name
            if form_name(form)['result']:
                # Form exists
                if form in forms:
                    print(f'The "{form}" already exist!')

                # Form not exists
                else:
                    break

            # Print the error
            else:
                print(form_name(form)['message'])

        # Begin the process
        try:
            print('Creating the new form...')

            # Form blueprint
            form_blueprint = f'{aurora_path + url_div}blueprints{url_div}form.zip'
            form_file = f'{app_path + url_div}forms{url_div + name + url_div}form.zip'
            form_dir = f'{app_path + url_div}forms{url_div + name + url_div}'
            
            # Copy the form blueprint
            copy_file(form_blueprint, form_file)

            # Unzip the form blueprint
            unzip_file(form_file, form_dir)

            # Delete the forms zip file
            delete_file(form_file)

            # Rename the _form.py
            old_form = f'{app_path + url_div}forms{url_div + name + url_div}_form.py'
            new_form = f'{app_path + url_div}forms{url_div + name + url_div + form}.py'
            rename_file(old_form, new_form)

            # Update the form blue print
            replace_file_string(new_form, 'FormName', form)

            # Update the _forms.py
            forms_file = f'{app_path + url_div}forms{url_div + name + url_div}_forms.py'

            forms_data = f"""    '{form}',\n"""
            forms_data += """)#do-not-change-me"""

            replace_file_line(forms_file, ')#do-not-change-me', forms_data)

            # Update the __init__.py
            init_file = f'{app_path + url_div}forms{url_div + name + url_div}__init__.py'

            if len(forms) == 0:
                replace_file_line(init_file, '...', '')

            init_data = f"""    from .{form} import {form}\n"""
            init_data += """    #do-not-change-me\n"""

            replace_file_line(init_file, '#do-not-change-me', init_data)

            # Print the result
            print('The new form created successfully!')
        
        # Handle errors
        except NameError as err:
            print(err)


    ##
    # @desc delete_form method for removing an existing form
    ##
    @click.command()
    def delete_form():
        # Prompt for app name
        while True:
            name = input("App Name: ")

            # Check app name
            if app_name(name)['result']:
                # App not exists
                if not name.lower() in apps:
                    print(f'The "{name}" doesn\'t exist!')

                # App exists
                else:
                    break

            # Print the error
            else:
                print(app_name(name)['message'])

        # Fetch registered forms for the app
        forms_module = importlib.import_module(f'forms.{name}._forms')
        forms = getattr(forms_module, "forms")
        
        # Prompt for form name
        while True:
            form = input("Form Name: ")

            # Check form name
            if form_name(form)['result']:
                # Form not exists
                if not form in forms:
                    print(f'The "{form}" doesn\'t exist!')

                # Form exists
                else:
                    break

            # Print the error
            else:
                print(form_name(form)['message'])
        
        # Alert the user for data loss
        alert = '''You will loose the following data perminantly:\n'''
        alert += f'''  {app_path + url_div}forms{url_div + name + url_div + form}.py\n'''
        
        # Print the alert
        print(alert)

        # Prompt the user for confirmation
        confirm = input("Do you want to continue? (yes/no) ")
            
        # Confirmed
        if confirm.lower() == 'yes':
            # Begin the process
            try:
                print('Removing the form...')

                # Delete the form module
                form_file = f'{app_path + url_div}forms{url_div + name + url_div + form}.py'
                delete_file(form_file)

                # Update the _forms.py
                forms_file = f'{app_path + url_div}forms{url_div + name + url_div}_forms.py'
                replace_file_line(forms_file, f"'{form}',", '')

                # Update the __init__.py
                init_file = f'{app_path + url_div}forms{url_div + name + url_div}__init__.py'

                if len(forms) == 1:
                    replace_file_line(init_file, f'from .{form} import {form}', '    ...\n')
                else:
                    replace_file_line(init_file, f'from .{form} import {form}', '')

                # Print the result
                print('The form deleted successfully!')

            # Handle errors
            except NameError as err:
                print(err)

        # Rejected
        else:
            print('The operation canceled!')
            exit()


    ##
    # @desc check_db method for checking the database for existence and so on
    ##
    @click.command()
    def check_db():
        # Check database existence
        if not db._exist_database(database):
            # Prepare the alert message
            alert = '''Database connection established successfully!\n'''
            alert += '''Database not initialized!\n'''
            alert += '''To initialize the database run the following command:\n'''
            alert += '''python manage.py init-db'''
            
            # Alert the user
            print(alert)

            # Exit the program
            exit()

        # Check the database migrations table
        if not db._exist_table('migrations'):
            # Prepare the alert message
            alert = '''Database not supported!\n'''
            alert += '''Make sure you initialized the database correctly, using the following command:\n'''
            alert += '''python manage.py init-db'''
            
            # Alert the user
            print(alert)

            # Exit the program
            exit()

        # Check the models for any errors
        if True:
            ...
            print('Checking models for any errors...')
            print('Work in progress.')

        # Check the available unmigrated changes
        if True:
            ...
            print('Checking models for unapplied changes...')
            print('Work in progress.')

        # Check the available migrations
        if True:
            ...
            print('Checking database for available migrations...')
            print('Work in progress.')

        # Exit the program
        exit()


    ##
    # @desc init_db method for initializing the database for the first time
    ##
    @click.command()
    def init_db():
        # db._delete_database(database=database, confirm=True)

        # Check database existence
        if db._exist_database(database):
            # Prepare the alert message
            alert = '''Database already exists!\n'''
            alert += '''To check the database run the following command:\n'''
            alert += '''python manage.py check-db'''
            
            # Alert the user
            print(alert)

        # Initialize the database
        else:
            print('Initializing the database...')

            # Attempt the process
            try:
                # Create the database
                db._create_database(database=database)

                # Create the migrations table
                new_db = Database()
                col_type = {
                    'id': 'integer',
                    'version': 'varchar(100)',
                    'current': 'boolean',
                    'date': 'date',
                    'comment': 'text',
                }
                primary_key = 'id'
                unique = ['version']
                not_null = ['version', 'current', 'date']
                default = {
                    'current': False,
                }
                new_db._create_table(table='migrations', col_type=col_type, primary_key=primary_key, unique=unique, not_null=not_null, default=default)

                # print()

                # Migration content placeholder
                migration_content = ''

                # Models collection for migrations
                models_coll = f"""_models = {models}\n\n"""

                # Update migrations content
                migration_content += models_coll

                # Create the users' tables (models)
                for model in models:
                    # Find the model and its class
                    Model = importlib.import_module(f'models.{model}')
                    Class = getattr(Model, model)

                    # The table name
                    table = Class().table

                    # Atributes dictionary
                    attrs = {}

                    # Add model columns (class attributes) to attrs
                    attrs.update(dict([(x,y) for x,y in Class.__dict__.items() if not x.startswith('__')]))

                    # Table default parameters
                    col_type = {}
                    primary_key = Class().primary_key
                    unique = []
                    not_null = []
                    default = {}
                    foreign_key = {}

                    # New attrs
                    new_attrs = {}

                    # Final attrs
                    final_attrs = {}

                    # Check primary key
                    if primary_key == None:
                        # Check the id column
                        if 'id' in attrs:
                            # Set primary key
                            primary_key = 'id'

                        else:
                            # Set primary key
                            primary_key = 'id'

                            # Add id to new attrs
                            new_attrs['id'] = {'datatype': 'integer', 'unique': False, 'not_null': True, 'default': None, 'check': None, 'foreign_key': None, 'on_update': None, 'on_delete': None}

                    # Update new attrs
                    new_attrs.update(attrs)

                    # Loop the model attributes
                    for x in new_attrs:

                        # Column + Datatype
                        col_type[x] = new_attrs[x]['datatype']

                        # UNIQUE
                        if new_attrs[x]['unique']:
                            unique.append(x)

                        # NOT NULL
                        if new_attrs[x]['not_null']:
                            not_null.append(x)

                        # DEFAULT
                        if new_attrs[x]['default'] != None:
                            default[x] = new_attrs[x]['default']

                        # Foreign key
                        if new_attrs[x]['related_to']:
                            # Find the reference model and its class
                            r_model = importlib.import_module(f"models.{new_attrs[x]['related_to']}")
                            r_class = getattr(r_model, new_attrs[x]['related_to'])

                            r_table = r_class().table
                            r_column = r_class().primary_key if r_class().primary_key else 'id'

                            foreign_key[x] = {
                                'r_table': r_table,
                                'r_column': r_column,
                                'on_update': new_attrs[x]['on_update'],
                                'on_delete': new_attrs[x]['on_delete'],
                            }

                    # Create the model table
                    new_db._create_table(table=table, col_type=col_type, primary_key=primary_key, unique=unique, not_null=not_null, default=default, foreign_key=foreign_key)

                    # Produce final attrs for migrations
                    final_attrs['table'] = table
                    final_attrs['col_type'] = col_type
                    final_attrs['primary_key'] = primary_key
                    final_attrs['unique'] = unique
                    final_attrs['not_null'] = not_null
                    final_attrs['default'] = default
                    final_attrs['foreign_key'] = foreign_key

                    # Model dictionary for migrations
                    model_dic = f"""{model} = {final_attrs}\n\n"""

                    # Update migrations content
                    migration_content += model_dic

                # Create the initial migration
                print('Creating the initial migration...')

                date = datetime.now().strftime("%Y-%m-%d")
                version = f'1-{date}'

                # Insert the initial migration to database
                new_db.create(table='migrations', data={'version':version,'current':True, 'date':date, 'comment':'The initial migration.'})
                
                # Create the migration file
                create_file(f'{app_path + url_div}migrations{url_div + version}.py', migration_content)
                # print(migration_content)

                # Print the message
                print('Database initialized successfully!')

            # Handle errors
            except NameError as err:
                print(err)

        # Exit the program
        exit()


    ##
    # @desc migrate_db method for migrating the database changes
    ##
    @click.command()
    def migrate_db():
        print('Work in progress!')




################
# Dependencies #
################
import os
import re
import string
import random
import shutil
from zipfile import ZipFile
from pathlib import Path


###################
# String Handling #
###################
##
# @desc Generates random string
# 
# @param [size]: int -- The size of output (characters)
# @param [chars]: str -- The type of characters
# 
# @return str
##
def random_string(size:int=8, chars:str=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(size))


##
# @desc Converts CamelCase to snake_case
#
# @param name: str - The name in CamelCase
#
# @retun str
##
def snake_case(name:str):
    return re.sub(r'(?<!^)(?=[A-Z])', '_', name).lower()


##
# @desc Delete characters from a string
#
# @param string: str - The string to format
# @param char: str - The character to delete in the string
#
# @retun str
##
def delete_chars(string:str, char:str):
    return re.sub(r'{}'.format(char), '', string)


################
# Collections  #
################
##
# @desc Convert tuple list into dictionary list, for SQLite Database
#
# @param cur: object - database connection cursor
# @param row: object - database rows
#
# @var d: dict
#
# @retun dict
##
def dict_factory(cur:object, row:object):
    d = {}

    for i, col in enumerate(cur.description):
        d[col[0]] = row[i]

    return d


##
# @desc Convert named list into dictionary list, for PostgreSQL Database
#
# @param cur: list - a named list
#
# @var translate: list - a dictionary list
#
# @retun list (of dictionaries)
##
def real_dict(cur:list):
    translate = []

    for x in cur:
        translate.append(dict(x))

    return translate


#################
# File Handling #
#################
##
# @desc Validates file name
#
# @param file_name: str - The file name to check
#
# @var regex: str - Regular expression
#
# @retun bool
##
def check_name(file_name:str):
    # Check required files
    if not file_name:
        return False

    # Regular expression
    regex = f'^[a-zA-Z]+[a-zA-Z_-]*$'

    # Valid name
    if re.match(regex, file_name):
        return True

    # Invalid name
    else:
        return False


##
# @desc Validates file name and extension
#
# @param src: str - The source file path (ex. 'example.py')
# @param extension: str - The file extension (ex. '.py', '.*')
# @param safe: bool - Safe character names (a-z, A-Z, _)
#
# @var regex: str - Regular expression
#
# @retun bool
##
def check_file(src:str, extension:str, safe:bool=True):
    # Check required files
    if not src or not extension:
        return False

    if extension == '.*':
        extension = '.[a-zA-Z]+'

    # Regular expression
    regex = f'^[a-zA-Z_]+[a-zA-Z0-9_]*\{extension}$' if safe else f'.*\{extension}$'

    # Valid name and extension
    if re.match(regex, src):
        return True

    # Invalid name or extension
    else:
        return False


##
# @desc Creates a file if not exists
#
# @param file: str - The absolute file path
# @param content: str - The content of the file
#
# @retun bool
##
def create_file(file:str, content:str=''):
    # File not exists
    if not os.path.exists(file):
        # Create the file
        f = open(file, 'x')
        f.write(content)
        f.close()

        # Return result
        return True

    # File already exists
    else:
        # Return Result
        return False


##
# @desc Writes to a file if exists
#
# @param file: str - The absolute file path
# @param content: str - The content of the file
#
# @retun bool
##
def write_file(file:str, content:str=''):
    # File already exists
    if os.path.exists(file):
        # Write to the file
        f = open(file, 'w')
        f.write(content)
        f.close()

        # Return result
        return True

    # File not exists
    else:
        # Return Result
        return False


##
# @desc Writes to a file if exists
#
# @param src: str - The source file path
# @param dst: str - The destination file path
#
# @retun bool
##
def copy_file(src:str, dst:str):
    # Source file exists
    if os.path.exists(src):
        # Copy the source file to the destination file
        shutil.copyfile(src, dst)

        # Return result
        return True

    # Source file not exists
    else:
        # Return Result
        return False


##
# @desc Writes to a file if exists
#
# @param src: str - The absolute source path
# @param dst: str - The absolute destination path
#
# @retun bool
##
def move_file(src:str, dst:str):
    # Source file exists
    if os.path.exists(src):
        # Check the file
        if re.search('[.]', src) and re.search('[.]', dst):
            # Copy the source file to the destination file
            shutil.move(src, dst)

            # Return result
            return True

        # Missing extension
        else:
            # Return Result
            return False

    # Source file not exists
    else:
        # Return Result
        return False


##
# @desc Writes to a file if exists
#
# @param src: str - The absolute source path
# @param dst: str - The absolute destination path
#
# @retun bool
##
def rename_file(src:str, dst:str):
    # Source file exists
    if os.path.exists(src):
        # Check the file
        if re.search('[.]', src) and re.search('[.]', dst):
            # Rename the source file to the destination file
            os.rename(src, dst)

            # Return result
            return True

        # Missing extension
        else:
            # Return Result
            return False

    # Source file not exists
    else:
        # Return Result
        return False


##
# @desc Removes a file if exists
#
# @param file: str - The absolute file path
#
# @retun bool
##
def delete_file(file:str):
    # File already exists
    if os.path.exists(file):
        # Remove the file
        os.remove(file)

        # Return result
        return True

    # File not exists
    else:
        # Return Result
        return False


##
# @descunzips a file to a directory
#
# @param file_path: str - The absolute zip file path
# @param dest_dir: str - The destination directory to unzip the file
#
# @retun bool -- True (on success), False (on error)
##
def unzip_file(file_path:str, dest_dir:str):
    with ZipFile(file_path, 'r') as zip_ref:
        zip_ref.extractall(dest_dir)


##
# @desc Replace strings in a file with new ones
#
# @param file_path: str - The absolute file path
# @param old_str: str - The old string
# @param new_str: str - The new string
#
# @retun bool -- True (on success), False (on error)
##
def replace_file_string(file_path:str, old_str:str, new_str:str):
    # Read the file
    with open(file_path, 'r') as file :
        f = file.read()

    # Replace the string
    f = f.replace(old_str, new_str)

    # Write the file out again
    with open(file_path, 'w') as file:
        file.write(f)


##
# @desc Replace lines in a file contain a string with new line data
#
# @param file_path: str - The absolute file path
# @param old_line: str - The character to match in the line
# @param new_str: str - The new line data
#
# @retun bool -- True (on success), False (on error)
##
def replace_file_line(file_path:str, old_line:str, new_line:str):
    # Open file
    with open(file_path, 'r+') as f:
        
        # Read lines
        lines = f.readlines()
        
        # Set the position to the beginning of the file 
        f.seek(0)

        # Loop the lines
        for line in lines:
            # Replace old lines
            if old_line in line:
                f.write(new_line)

            # Write other lines
            else:
                f.write(line)

        # Resize the file to the current file stream position
        f.truncate()


######################
# Directory Handling #
######################
##
# @desc Checks if a directory exists
#
# @param dir: str - The directory string
#
# @retun bool -- True (directory exists), False (direcory does not exist)
##
def dir_exist(dir:str):
    # Directory exists
    if os.path.isdir(dir):
        return True

    # Directory doesn't exist 
    else:
        return False


##
# @desc Checks if a directory is empty
#
# @param dir: str - The directory string
# 
# @var dir_list: list - Lists the direcotry sub direcories and files
#
# @retun bool -- True (directory is empty), False (direcory is not empty)
##
def dir_empty(dir:str):
    # Directories list
    dir_list = os.listdir(dir)

    # Directory is empty
    if len(dir_list) == 0:
        return True

    # Directory is not empty
    else:    
        return False


##
# @desc Create a directory if not exists
#
# @param dir: str - The directory string
#
# @retun bool -- True (on success), False (on error)
##
def make_dir(dir:str):
    # Try to create the directory
    try:
        Path(dir).mkdir(parents=True, exist_ok=True)
        return True
    
    # Handle errors
    except:
        return False


##
# @desc removes a directory and all its contents
#
# @param dir: str - The directory string
#
# @retun bool -- True (on success), False (on error)
##
def delete_dir(dir:str):
    # Try to create the directory
    try:
        shutil.rmtree(dir)
        return True
    
    # Handle errors
    except:
        return False


#################
# CLI Helpers #
#################
##
# @desc Validates app name
#
# @param name: str - The app name
#
# @retun dict
##
def app_name(name:str):
    # Check required app name
    if not name:
        return {
            'result': False, 
            'message': 'The app name is required!'
        }

    # Regular expression
    regex = '^[a-z_]+$'

    # Valid name
    if re.match(regex, name):
        return {
            'result': True, 
            'message': ''
        }

    # Invalid name
    else:
        return {
            'result': False, 
            'message': 'The app name is invalid.\nValid characters: a-z, _'
        }


##
# @desc Validates base url
#
# @param url: str - The app name
#
# @retun dict
##
def base_url(url:str):
    # Check required base URL
    if not url:
        return {
            'result': False, 
            'message': 'The base URL is required!'
        }

    # Regular expression
    regex = '^[a-z-]+$'

    # Valid URL
    if re.match(regex, url):
        return {
            'result': True, 
            'message': ''
        }

    # Invalid URL
    else:
        return {
            'result': False, 
            'message': 'The base URL is invalid.\nValid characters: a-z, -'
        }


##
# @desc Validates controller name
#
# @param name: str - The controller name
#
# @retun dict
##
def controller_name(name:str):
    # Check required controller name
    if not name:
        return {
            'result': False, 
            'message': 'The controller name is required!'
        }

    # Regular expression
    regex = '^(?:[A-Z][a-z]+)+$'

    # Valid name
    if re.match(regex, name):
        return {
            'result': True, 
            'message': ''
        }

    # Invalid name
    else:
        return {
            'result': False, 
            'message': 'The controller name must be in "CamelCase" form with at least two "a-z" and "A-Z" characters.'
        }


##
# @desc Validates controller url
#
# @param url: str - The app name
#
# @retun dict
##
def controller_url(url:str):
    # Optional base URL
    if not url:
        return {
            'result': True, 
            'message': ''
        }

    # Regular expression
    regex = '^[a-z]+[a-z-<>:/]*$'

    # Valid URL
    if re.match(regex, url):
        return {
            'result': True, 
            'message': ''
        }

    # Invalid URL
    else:
        return {
            'result': False, 
            'message': 'The controller URL is invalid.\nValid characters: a-z, -, /, <, :, >'
        }


##
# @desc Validates controller methods
#
# @param methods: list - The controller methods
#
# @retun dict
##
def controller_methods(methods:list):
    # Optional methods
    if not methods:
        return {
            'result': True, 
            'message': ''
        }

    # Valid methods
    valid_methods = ['POST', 'PUT', 'GET', 'DELETE']

    # Validate methods
    for method in methods:
        # Invalid methods
        if not method.upper() in valid_methods:
            return {
                'result': False, 
                'message': 'Valid Methods: POST, GET, PUT, DELETE'
            }

        # Valid name
        else:
            return {
                'result': True, 
                'message': ''
            }


##
# @desc Validates view name
#
# @param view: str - The view name
#
# @retun dict
##
def view_name(view:str):
    # Check required base URL
    if not view:
        return {
            'result': False, 
            'message': 'The view is required!'
        }

    # Regular expression
    regex = '^[a-z-_]+$'

    # Valid view
    if re.match(regex, view):
        return {
            'result': True, 
            'message': ''
        }

    # Invalid view
    else:
        return {
            'result': False, 
            'message': 'The view is invalid.\nValid characters: a-z, -, _'
        }


##
# @desc Validates model name
#
# @param name: str - The model name
#
# @retun dict
##
def model_name(name:str):
    # Check required model name
    if not name:
        return {
            'result': False, 
            'message': 'The model name is required!'
        }

    # Regular expression
    regex = '^(?:[A-Z][a-z]+)+$'

    # Valid name
    if re.match(regex, name):
        return {
            'result': True, 
            'message': ''
        }

    # Invalid name
    else:
        return {
            'result': False, 
            'message': 'The model name must be in "CamelCase" form with at least two "a-z" and "A-Z" characters.'
        }


##
# @desc Validates form name
#
# @param name: str - The form name
#
# @retun dict
##
def form_name(name:str):
    # Check required form name
    if not name:
        return {
            'result': False, 
            'message': 'The form name is required!'
        }

    # Regular expression
    regex = '^(?:[A-Z][a-z]+)+$'

    # Valid name
    if re.match(regex, name):
        return {
            'result': True, 
            'message': ''
        }

    # Invalid name
    else:
        return {
            'result': False, 
            'message': 'The form name must be in "CamelCase" form with at least two "a-z" and "A-Z" characters.'
        }




################
# Dependencies #
################
import sys
import pathlib
import importlib
from .helpers import snake_case
from .SQL import Database


###############
# Model Class #
###############
##
# @desc Instantiates the Database class of the Database API
##
class Model:
    ##
    # @desc Constructor method
    #
    # @property table: str -- The table name of the model
    #
    # @var caller_path: str -- The caller path
    # @var caller_file: str -- The caller file
    # @var caller_name: str -- The caller name (the model name)
    # @var Model: module -- The model module
    # @var Class: type -- The model class
    # @var attrs: dict -- The model class attributes
    ##
    def __init__(self):
        # Find information about the caller
        caller_path = sys._getframe().f_back.f_code.co_filename
        caller_file = pathlib.PurePath(caller_path).name
        caller_name = caller_file.replace('.py', '')

        # The table name
        self.table = snake_case(caller_name)

        # The model primary key
        self.primary_key = None
        
        # Instantiate the Database class
        Database.__init__(self)


    ##
    # @desc Destructor method
    ##
    def __del__(self):
        # Destruction method
        Database.__del__(self)


    ##
    # @desc column method - for generating sql column and its data
    # 
    # @param datatype: str -- *Required column datatype
    # @param primary_key: bool -- Optional column PRIMARY KEY constraint (False by default)
    # @param unique: bool -- Optional column UNIQUE constraint (False by default)
    # @param not_null: bool -- Optional column NOT NULL constraint (False by default)
    # @param default: any -- Optional column DEFAULT constraint (None by default)
    # @param check: str -- Optional column CHECK constraint (None by default)
    # @param foreign_key: str -- Optional column FOREIGN KEY constraint (None by default)
    # 
    # @var result: dict -- The result dictionary
    #
    # @return dict
    ##
    @staticmethod
    def column(datatype:str, unique:bool=False, not_null:bool=False, default:any=None, check:str=None, 
        related_to:str=None, on_update:str='CASCADE', on_delete:str='CASCADE'):

        # Import the configuration file
        config = importlib.import_module('config')

        # Debug mode
        debug = getattr(config, 'DEBUG')

        # Check required params
        if not datatype:
            # Check debug mode
            if debug:
                # Raise error
                raise Exception("You must provide the required parameters: ['datatype']")

            # Return the result
            else:
                print('You must provide the required parameters.')
                return False

        # Result dictionary placeholder
        result = {}
        
        result['datatype'] = datatype
        result['unique'] = True if unique == True else False
        result['not_null'] = True if not_null == True else False
        result['default'] = default if default != None else None
        result['check'] = check if check else None

        # Foreign key
        result['related_to'] = related_to if related_to else None
        result['on_update'] = on_update if related_to and on_update else None
        result['on_delete'] = on_delete if related_to and on_delete else None

        return result


    ##
    # @desc query method
    #
    # @param sql: str -- *Required SQL statement (ex. "SELECT * FROM users")
    #
    # @var result: object -- Database query
    #
    # @return any -- Query result
    ##
    def query(self, *class_args, **class_kwargs):
        return Database.query(self, *class_args, **class_kwargs)


    ##
    # CAUTION! Use this methods only in developement.
    #
    # @desc Checks if a database (file - SQLite) exists
    #
    # @param database: str -- *Required database name (file - SQLite)
    #
    # @return bool -- Query result
    ##
    def _exist_database(self, *class_args, **class_kwargs):
        return Database._exist_database(self, *class_args, **class_kwargs)


    ##
    # @desc Insert single row
    #
    # @param data: dict -- *Required data (ex. {"username": "john-doe", "password": "123456"})
    # 
    # @return int -- Last inserted id as query result on success
    # @return False: bool -- As query result on error
    ##
    def create(self, data:dict):
        return Database.create(self, table=self.table, data=data)


    ##
    # @desc Insert multi rows
    #
    # @param table: str -- *Required Table name (ex. "users")
    # @param data: list -- *Required data (ex. [{...}, {...}, ...])
    #
    # @return int -- Last inserted id
    ##
    def create_multi(self, data:list):
        # Loop data
        i = 1
        for x in data:
            if i == len(data):
                return self.create(data=x)

            else:
                self.create(data=x)

            i += 1


    ##
    # @desc Select rows
    #
    # @param cols: list -- Optional Columns (ex. ["id", "first_name", "last_name"])
    # @param where: dict -- Optional WHERE statement (ex. {"id": "2", "username": "admin"})
    # @param order_by: dict -- Optional ORDER BY statement (ex. {"id": "ASC", "date": "DESC"})
    # @param limit: int -- Optional LIMIT statement (ex. "10")
    #
    # @return class: type -- Query result
    ##
    def read(self, cols:list=[], where:dict={}, order_by:dict={}, group_by:str=None, limit:int=None, offset:int=None):
        return Database.read(self, table=self.table, cols=cols, where=where, order_by=order_by, group_by=group_by, 
                              limit=limit, offset=offset)


    ##
    # *CAUTION! If you ignore the 'where' parameter, it may updates the columns for all records!
    #
    # @desc Update row(s)
    #
    # @param data: dict -- *Required data (ex. {"first_name": "John", "last_name": "Doe"})
    # @param where: dict -- Optional (*CAUTION!) WHERE statement (ex. {"id": "2", "username": "admin"})
    #
    # @return bool -- Query result
    ##
    def update(self, data:dict, where:dict={}, confirm:bool=False):
        return Database.update(self, table=self.table, data=data, where=where, confirm=confirm)


    ##
    # *WARNING! If you ignore the 'where' parameter, it will deletes all the records inside your table! (must confirm)
    #
    # @desc Delete row(s)
    #
    # @param table: str -- *Required Table name (ex. "users")
    # @param where: dict -- Optional (*WARNING!) WHERE statement (ex. {"id": "2", "username": "admin"})
    # @param confirm: bool -- Required|Optional confirm (if not where it will be Required)
    #
    # @return bool -- Query result
    ##
    def delete(self, where:dict={}, confirm:dict=False):
        return Database.delete(self, table=self.table, where=where, confirm=confirm)




################
# Dependencies #
################
import importlib
from flask import Flask


################
# Aurora Class #
################
##
# @desc Aurora class to serve the root application and it's child apps 
##
class Aurora():

    ##
    # @desc Constructor method
    #
    # @property app: NoneType -- Placeholder for the root app (Flask Instance)
    # @property config: module -- The config module
    # @property debug: bool -- The debug mode
    # @property apps: module -- The _apps module
    ##
    def __init__(self):
        # The root app
        self.app = None
        
        # Import the config module
        self.config = importlib.import_module("config")

        # Debug mode
        self.debug = getattr(self.config, "DEBUG")
        
        # Import the _apps module
        self.apps = importlib.import_module("_apps")

        # Try to serve the root app
        try:
            self.serve()
        
        # Except errors
        except NameError as err:
            # Check debug mode
            if self.debug:
                # Raise error
                raise Exception(err)


    ##
    # @desc The serve method -- Serves the root application and its child apps
    #
    # @var apps: dict -- The APPS attribute of the _apps module
    # @var root_path: str -- The root app path
    # @var default_app: str -- The DEFAULT_APP attribute of the config module
    # @var statics: str -- The STATICS attribute of the config module
    # @var secret_key: str -- The SECRET_KEY attribute of the config module
    # @var globals: dict -- The GLOBALS attribute of the config module
    # @var supported_apis: list -- The supported database APIs for the selected database engine
    # @var error: str -- The error message on error
    # @var app: object -- The root application
    #
    # @decorator context_processor: function -- For decorating global_variables local method
    # @method global_variables: function -- A local method for setting global variables
    #
    # @return object -- The root app
    ##
    def serve(self):
        # Fetch the required attributes
        apps = getattr(self.apps, "apps")
        root_path = getattr(self.config, "ROOT_PATH")
        statics = getattr(self.config, 'STATICS')
        secret_key = getattr(self.config, "SECRET_KEY")

        # Initialize the root app (Flask instance)
        self.app = Flask(__name__, template_folder=f'{root_path}/views', static_folder=f'{root_path}/{statics}')
        
        # Set the app secret key
        self.app.config['SECRET_KEY'] = secret_key


        ##
        # @desc The local global_variables method -- Sets the global variables usable in views
        #
        # @var global_dict: dict -- A dictionary for all global variables (GLOBALS + Auto Globals)
        # @var translate dict -- Uppercase key version of global_dict
        #
        # @return dict -- Global variables dictionary
        ##
        app = self.app
        @app.context_processor
        def global_variables():
            translate = {}
            auto_globals = {}

            # Global attibutes
            error_app = getattr(self.config, 'ERROR_APP')
            default_app = getattr(self.config, 'DEFAULT_APP')
            globals = getattr(self.config, 'GLOBALS')
            
            # Auto globals
            auto_globals = {
                'error_app': apps[error_app],
                'default_app': apps[default_app],
                'statics': statics,
            }

            # Add apps
            auto_globals.update(apps)

            # Add global variables
            auto_globals.update(globals)

            # Translate auto_globals to uppercase
            for k, v in auto_globals.items():
                translate[k.upper()] = v
            
            # Return the translated dictionary
            return translate

        # Try to bootstrap the apps
        try:
            # Bootstrap installed apps (child apps)
            self.bootstrap(apps)

            # Return the root app
            return self.app
        
        # Handle errors
        except NameError as err:
            # Check debug mode
            if self.debug:
                # Raise error
                raise Exception(err)

            # Return the result
            else:
                return False


    ##
    # @desc The bootstrap method -- Bootsraps the child apps modules and packages
    #
    # @param apps: dict -- The child apps to serve
    # @param default_app: string -- The default app to serve the '/' url
    #
    # @var module: module -- The _controllers module
    # @var controllers: list -- The controllers attribute of the _controllers module
    #
    # @return bool
    ##
    def bootstrap(self, apps:dict):
        # Check the child apps
        if apps:
            # Import app modules and packages
            for app in apps:
                # Import the _controllers module
                module = importlib.import_module(f"controllers.{app}._controllers")

                # Fetch the controllers attribute
                controllers = getattr(module, "controllers")
                
                # Try to call the router method
                try:
                    self.router(app, controllers, apps)

                # Something went wrong
                except NameError as err:
                    # Check debug mode
                    if self.debug:
                        # Raise error
                        raise Exception(err)

                    # Return the result
                    else:
                        return False
            
            # Everything is OK
            return True


    ##
    # @desc The app router method -- Routes a controller for each route(url)
    #
    # @param app: str -- The child app to route
    # @param controllers: list -- The child app controllers
    # @param apps: dict -- The child apps to route
    # @param default_app: string -- The default app to route
    #
    # @var error_app: str -- The ERROR_APP attribute of the config module
    # @var default_app: str -- The DEFAULT_APP attribute of the config module
    # @var Module: module -- The controller module
    # @var Controller: class -- The controller class
    # @var methods: list -- ['GET'], ['POST'], ['GET', 'POST'] -- default: ['GET']
    # @var rule: str -- URL rule for a route
    # @var endpoint: str -- Endpoint for a route
    # 
    # @return None
    ##
    def router(self, app:str, controllers:list, apps:dict) -> None:
        error_app = getattr(self.config, 'ERROR_APP')
        default_app = getattr(self.config, 'DEFAULT_APP')

        for controller in controllers:

            # Recognize the controller
            Module = importlib.import_module(f"controllers.{app}.{controller[0]}")
            Controller = getattr(Module, controller[0])

            # Check for the mothods
            if len(controller) == 3:
                methods=controller[2]
            else:
                methods=['GET']

            # Check the URL rule
            if controller[1] == '':
                rule = f'/{apps[app]}/'
                endpoint = app
            else:
                rule = f'/{apps[app]}/{controller[1]}/'
                endpoint = f'{app}-{controller[1]}'

            # Generate the view function
            view_func = Controller.as_view(endpoint)

            # Check the apps
            # Error pages app
            if app == error_app:
                # Errors on abort
                self.app.register_error_handler(int(controller[1]), Controller.err)

                # Errors on GET method for developer
                if self.debug:
                    self.app.add_url_rule(rule=rule, endpoint=endpoint, view_func=view_func, methods=methods)

            # Other apps
            else:

                # Root app index controller
                if app == default_app and controller[1] == '':
                    self.app.add_url_rule(rule='/', endpoint='default-app', view_func=view_func, methods=methods)

                # All apps controllers
                self.app.add_url_rule(rule=rule, endpoint=endpoint, view_func=view_func, methods=methods)


    ##
    # @desc The run method -- Runs the root app
    # 
    # @var host: str -- The HOST attribute of the config module
    # @var port: str -- The PORT attribute of the config module
    # 
    # @return object: NoneType -- The root app
    ##
    def run(self):    
        # Fetch the required attributes
        host = getattr(self.config, "HOST")
        port = getattr(self.config, "PORT")

        # Try to run the app
        try:
            return self.app.run(host=host, port=port, debug=self.debug)

        except NameError as err:
            # Check debug mode
            if self.debug:
                # Raise error
                raise Exception(err)

            # Return the result
            else:
                return False

